;;
;; Komposter player (c) 2010 Firehawk/TDA
;;
;; This code is licensed under the MIT license:                             
;; http://www.opensource.org/licenses/mit-license.php
;;
;; This is a reference implementation for a Komposter playroutine
;; in x86 assembly language.
;;
;; For readability, this player has not been optimized for size and
;; for using it in an intro, it's recommended to comment out any code
;; that is not required by the song. Remember to also set any unused
;; jump table entries to zero for improved compressibility.
;;


; macros SONGDATA(address) and SONGBSS(address) are used for
; getting a pointer address to the data. change these macros
; if you relocate them
%define	SONGDATA(X) [X]
%define SONGBSS(X)  [X]

; output sample rate
%define 	OUTPUTFREQ	44100

; how many samples of audio we're rendering
%define		SONG_BUFFERLEN	20*OUTPUTFREQ

; flags in voice data byte
%define		FLAG_GATE		1
%define		FLAG_TRIG		2
%define		FLAG_ACCENT		4
%define		FLAG_NOTE		8
%define		FLAG_RESTART_ENV	16
%define		FLAG_RESTART_VCO	32
%define		FLAG_RESTART_LFO	64
%define		FLAG_LOAD_PATCH		128

; vco waveforms
%define		VCO_PULSE		0
%define		VCO_SAW			1
%define		VCO_TRIANGLE		2
%define		VCO_SINE		3

%define		MAX_MODULES		64  ; with dword values, this means 256 bytes per channel

%define 	DELAYBUFFERSIZE		262144 ; 18 bit offsets to delaybuffer


;
; DATA
;
section .data

outputfreq	dd	44100.0  ; outputfreq as a float value
noise_x1	dd	0x67452301
noise_x2	dd	0xefcdab89
noise_div	dd	4294967296.0
midi0           dd      8.1757989156 ; C0
midisemi        dd      1.059463094 ; ratio between notes a semitone apart

; small constants used around the code
half		dd	0.5
threefourths	dd	0.75
four		dd	4.0
five		dd	5.0

; these are for the 24db/oct lpf, comment out if not used
two_point_threetwo      dd      2.32
point_onefive           dd      0.15
lpf_feedback_coef       dd      0.35013  
point_three             dd      0.3



; jump table to module functions. if you know that your song doesn't use
; some module, delete the function from modules.asm and set the jump table
; address here to zero. probably saves a few bytes.
modfunctable:
	dd	module_func_kbd
	dd	module_func_env
	dd	module_func_vco
	dd	module_func_lfo
	dd	module_func_cv
	dd	module_func_amp
	dd	module_func_mixer
	dd	module_func_vcf
	dd	module_func_lpf24
	dd	module_func_delay
	dd	module_func_att
	dd	module_func_resample
	dd	module_func_aswitch
	dd	module_func_dist
	dd	module_func_accent
	dd	module_func_output

; include the song itself from an external file
%include "song.inc"



;
; BSS
;
section .bss

tick		resd	1
oldtick		resd	1
seqptr		resd	1

tempfloat	resd	1
sample		resd	1

moddata		resd	NUM_CHANNELS*MAX_MODULES*32 ; output,mod,moddata*16,reserved*14

pitch		resd	NUM_CHANNELS
flags		resb	NUM_CHANNELS

global songbuffer
songbuffer	resw	SONG_BUFFERLEN

patternpos	resd	NUM_CHANNELS
noteduration	resd	NUM_CHANNELS
seq_pattstart  	resd	NUM_CHANNELS
seq_pattptr	resd	NUM_CHANNELS
seq_rep		resb	NUM_CHANNELS
seq_trans	resb	NUM_CHANNELS
seq_patchptr	resd	NUM_CHANNELS

delaycount	resb	1
delaybuffer	resd	NUM_DELAYS*DELAYBUFFERSIZE



;
; TEXT
;
section .text

; code for the synth modules
%include "modules.asm"

; this function renders the actual audio into the songbuffer in BSS
; with SONG_BUFFERLEN samples 
global render_song
render_song:
	; starting playback, so initialize any variables that need to be
	; initialized. usually everything in the bss is zero, you'll just
	; need to set any nonzero
	dec 	byte SONGBSS(tick)
	lea	eax, SONGDATA(seqevents)
	mov	SONGBSS(seqptr), eax ; set seq pointer to start of eventlist

	; start the loop to fill the outputbuffer with rendered audio
	xor	ebx, ebx ; ebx = sample number
.sample_loop:
	push	ebx ; save sample number

	mov	eax, SONGBSS(tick)
	mov	SONGBSS(oldtick), eax
	mov	eax, ebx
	xor	edx, edx
	idiv	dword SONGDATA(tickdivider)  ; tickdivider = OUTPUTFREQ/(bpm*256/60)
	mov	SONGBSS(tick), eax	; ticks is now in eax

	; process seq eventlist
	mov	esi, SONGBSS(seqptr)
.seq_checkevent:
	shr 	eax, 6
	cmp 	ax, [esi]
	jb 	near .seq_noevent
.seq_doevent:
	add 	esi, 2
	lodsb
	movzx 	edx, al
	xor 	eax, eax
	mov	SONGBSS(patternpos+edx), eax ; patternpos to 0
	lodsb
	mov	SONGBSS(seq_rep+edx), al ; set repeat
	lodsb
	movzx	ebx, word SONGDATA(patternstart+eax*2)
	lea	ebx, SONGDATA(patterndata+ebx) ; actual offset in memory
	mov	SONGBSS(seq_pattstart+edx*4), ebx ; set pattern start
	mov	SONGBSS(seq_pattptr+edx*4), ebx ; pattern playposition
	lodsb
	mov	SONGBSS(seq_trans+edx), al ; set transpose
        ; load patch to synth
	lodsb
	mov	ax, word SONGDATA(patchstart+eax*2) ; start of new patch data
	lea	eax, SONGDATA(patchdata+eax*4)
	mov	SONGBSS(seq_patchptr+edx*4), eax
	or	byte SONGBSS(flags+edx), FLAG_LOAD_PATCH

	;event is done, check if there's more
	mov	SONGBSS(seqptr), esi ; set seqptr to next event
	and	byte SONGBSS(flags+edx), 0xff-FLAG_NOTE ; play the first note next
	mov	eax, SONGBSS(tick)
	jmp	.seq_checkevent
.seq_noevent:

	; now loop through all the voices for this sample and render
	; a mixed output into xmm7
	xor	edx, edx ; edx=voice
	mov	SONGBSS(sample), edx ; set mixed sample to zero
.voice_loop:
	; play the pattern on this voice
	mov	eax, SONGBSS(tick)
	cmp	eax, SONGBSS(oldtick)
	jz	near .mod_process
	and 	al, 63
	jz	.pattern_tick0
	cmp	al, 60
	jnz	near .mod_process
	; tick 60/124/188/252,
	cmp	byte SONGBSS(noteduration+edx), 0
	jz	.pattern_gateoff
	dec	byte SONGBSS(noteduration+edx) ; yep, decrease the duration
	jmp	.mod_process
.pattern_gateoff:
	and	byte SONGBSS(flags+edx), 0xff-(FLAG_GATE+FLAG_NOTE) ; drop gate and read next note
        jmp	.mod_process

.pattern_tick0:
	; on tick 0/64/128/192, process notes and rests
	test	byte SONGBSS(flags+edx), FLAG_NOTE
	jnz	near .mod_process
.pattern_newnote: ; read a new note from the pattern
	mov	esi, SONGBSS(seq_pattptr+edx*4)
	and	esi, esi
	jz 	near .mod_process
.pattern_read:
	lodsb ; al = byte from pattern
	cmp	al, 0xff
	jnz	.patt_play
	; end of pattern, decrease repeat count
	cmp	byte SONGBSS(seq_rep+edx), 0
	jz	.patt_done ; all repeats done, skip to end
	dec	byte SONGBSS(seq_rep+edx)
	mov	eax, SONGBSS(seq_pattstart+edx*4)
	mov	SONGBSS(seq_pattptr+edx*4), eax
	jmp	.pattern_newnote ; read the first byte right away
.patt_play:
	; byte was not 0xff, so play it
	or	byte SONGBSS(flags+edx), FLAG_NOTE
	mov	SONGBSS(noteduration+edx), al
	and	byte SONGBSS(noteduration+edx), 0x7f ; strip note bit
	test	al, 80h ; was this a note?
	jz	.patt_notedone
	; it was a note, load the next byte
	lodsb
	and	byte SONGBSS(flags+edx), 0xff-FLAG_ACCENT ; clear accent
	test	al, 80h ; accented note?
	jz	.patt_noaccent
	or	byte SONGBSS(flags+edx), FLAG_ACCENT
	and	al, 0x7f
.patt_noaccent:
	add	al, SONGBSS(seq_trans+edx) ; transpose
	; set pitch according to midi note
	fld	dword SONGDATA(midi0)
.patt_setnote_mul:
	fmul	dword SONGDATA(midisemi)
	dec	al
	jnz	.patt_setnote_mul
	fstp	dword SONGBSS(pitch+edx*4)
	; finally, set gate, trig and all restarts
	mov 	al, FLAG_GATE|0xf0
	and	al, SONGDATA(seqmask+edx) ; mask out restart bits
	or	byte SONGBSS(flags+edx), al
.patt_notedone:
	mov	SONGBSS(seq_pattptr+edx*4), esi ; store esi back
.patt_done:

	; process the signal stack by looping through all the
	; modules
.mod_process:
	movzx	ebp, byte SONGDATA(seqvoice+edx)
	movzx	ebp, byte SONGDATA(synthstart+ebp*2)
	; ebp = ptr to synth start, index to synth data

	xor	ecx, ecx ;ecx = module index
	mov	edi, edx ; edx=channel
	shl 	edi, 13 ; 8192 bytes per channel, 128 bytes per module
	lea 	edi, SONGBSS(moddata+edi) ; edi = start of moddata for this channel
	mov	esi, edi
.mod_process_loop:
	; load a new patch?
	test	byte SONGBSS(flags+edx), FLAG_LOAD_PATCH
	jz	.mod_process_prepare
	; load value from patch to modulator
	mov	eax, SONGBSS(seq_patchptr+edx*4)
	mov	eax, [eax+ecx*4]
	mov	[edi+4], eax

	; store modulator first to fpu
.mod_process_prepare:
	fninit
	fld	dword [edi+4] ; modulator at +4

	; collect input signal voltages to fpu
	pushad
	add 	ebp, ecx 
	mov 	eax, SONGDATA(modinputs+ebp*4)
	mov	ecx, 4
.mod_signal_loop:
	xor	ebx, ebx
	mov	bl, al ; ebx = input module index
	shl	ebx, 7
	fld 	dword [esi+ebx] ; output at offset 0
	shr	eax, 8 ; next byte to al
	loop	.mod_signal_loop
	popad

	movzx	ebx, byte SONGDATA(modtypes+ebp+ecx) ; ebx=module type [0..127]
	pushad
	mov 	eax, [edi+4]		; eax = modulator as integer
	mov	cl, SONGBSS(flags+edx)  ; cl = flags
	lea 	esi, [edi+8]            ; esi = data area starts at +8
	call 	SONGDATA(modfunctable+ebx*4)
	popad
	fst 	dword [edi] ; store output

	; update noise generator on each module loop
	mov	eax, SONGBSS(noise_x1)
	xor	eax, SONGBSS(noise_x2)
	add	SONGBSS(noise_x2), eax
	mov	SONGBSS(noise_x1), eax

	; next module
	inc	ecx
	add 	edi, 128
	cmp	bl, 0fh
	jnz	.mod_process_loop
.mod_process_end:
	and	byte SONGBSS(flags+edx), 0x0f ; clear hard restart flags from flags

	fadd	dword SONGBSS(sample) ; mix previous channels in
	fstp	dword SONGBSS(sample) ; and pop back to temp storage
	inc	edx			; next voice
	cmp	edx, NUM_CHANNELS
	jl	.voice_loop
	; all voices done, sample now has the final channel mix

	pop	ebx ; restore sample number
	fld	dword SONGBSS(sample)
	fmul	dword SONGDATA(samplemul)
	fistp	word SONGBSS(songbuffer+ebx*2)
	inc	ebx
	cmp	ebx, SONG_BUFFERLEN
	jl	.sample_loop

	;done, the buffer is now full of music. :)
	ret

playercode_end:

;;
;; eof
;;